use std::{
    collections::HashMap,
    path::{Path, PathBuf},
    process::Command,
};

use ambient_driver::action::{TrustedAction, UnsafeAction};
use serde::{Deserialize, Serialize};
use tempfile::tempdir;

use crate::{adminlog::AdminLog, AdapterError};

// Exit code to indicate we didn't get one from the process.
const NO_EXIT: i32 = 999;

// Read `.radicle/ambient.yaml` from the source tree.
#[derive(Debug, Default, Deserialize, Serialize)]
#[serde(deny_unknown_fields)]
pub struct DistilledPlan {
    pre_plan: Option<Vec<TrustedAction>>,
    plan: Option<Vec<UnsafeAction>>,
    post_plan: Option<Vec<TrustedAction>>,
}

impl DistilledPlan {
    pub fn read(filename: &Path) -> Result<Self, AdapterError> {
        let data =
            std::fs::read(filename).map_err(|err| AdapterError::ReadPlan(filename.into(), err))?;
        serde_yml::from_slice(&data).map_err(|err| AdapterError::ParsePlan(filename.into(), err))
    }
}

#[derive(Debug, Serialize)]
pub struct Project {
    source: PathBuf,
    image: PathBuf,
    pre_plan: Vec<TrustedAction>,
    plan: Vec<UnsafeAction>,
    post_plan: Vec<TrustedAction>,
    artifact_max_size: Option<u64>,
    cache_max_size: Option<u64>,
}

#[derive(Debug, Serialize)]
pub struct Projects {
    projects: HashMap<String, Project>,
}

impl Projects {
    pub fn write(&self, filename: &Path) -> Result<String, AdapterError> {
        let text = serde_yml::to_string(&self).map_err(AdapterError::ToYaml)?;
        std::fs::write(filename, text.as_bytes()).map_err(AdapterError::WriteProjects)?;
        Ok(text)
    }

    pub fn project_name(&self) -> String {
        let keys: Vec<&str> = self.projects.keys().map(|s| s.as_str()).collect();
        assert_eq!(keys.len(), 1);
        keys[0].to_string()
    }
}

#[derive(Debug, Default)]
pub struct ProjectsBuilder {
    name: Option<String>,
    source: Option<PathBuf>,
    image: Option<PathBuf>,
    pre_plan: Vec<TrustedAction>,
    plan: Vec<UnsafeAction>,
    post_plan: Vec<TrustedAction>,
    artifact_max_size: Option<u64>,
    cache_max_size: Option<u64>,
}

impl ProjectsBuilder {
    pub fn name(mut self, name: &str) -> Self {
        self.name = Some(name.into());
        self
    }

    pub fn source(mut self, path: &Path) -> Self {
        self.source = Some(path.into());
        self
    }

    pub fn image(mut self, path: &Path) -> Self {
        self.image = Some(path.into());
        self
    }

    pub fn artifact_max_size(mut self, size: u64) -> Self {
        self.artifact_max_size = Some(size);
        self
    }

    pub fn cache_max_size(mut self, size: u64) -> Self {
        self.cache_max_size = Some(size);
        self
    }

    pub fn plan(mut self, plan: &DistilledPlan) -> Self {
        if let Some(actions) = &plan.pre_plan {
            self.pre_plan = actions.to_vec();
        }
        if let Some(actions) = &plan.plan {
            self.plan = actions.to_vec();
        }
        if let Some(actions) = &plan.post_plan {
            self.post_plan = actions.to_vec();
        }
        self
    }

    pub fn build(self) -> Projects {
        let projects: HashMap<String, Project> = HashMap::from([(
            self.name.unwrap(),
            Project {
                source: self.source.expect("ProjectsBuilder::source is set"),
                image: self.image.expect("ProjectsBuilder::image is set"),
                pre_plan: self.pre_plan.clone(),
                plan: self.plan.clone(),
                post_plan: self.post_plan.clone(),
                artifact_max_size: self.artifact_max_size,
                cache_max_size: self.cache_max_size,
            },
        )]);
        Projects { projects }
    }
}

pub fn run_ambient_driver(
    adminlog: &mut AdminLog,
    dry_run: bool,
    projects: &Projects,
    cpus: usize,
    memory: usize,
    state: &Path,
) -> Result<(i32, Option<Vec<u8>>), AdapterError> {
    let tmp = tempdir().map_err(AdapterError::TempDir)?;
    let projects_filename = tmp.path().join("projects.yaml");

    let project_name = projects.project_name();
    let projects = projects.write(&projects_filename)?;
    indented(adminlog, "projects", projects.as_bytes());

    let config_filename = tmp.path().join("config.yaml");
    std::fs::write(
        &config_filename,
        format!(
            r#"---
state: "{}"
executor: "/usr/bin/ambient-execute-plan"
cpus: {cpus}
memory: "{memory}"
...
"#,
            state.display()
        ),
    )
    .map_err(AdapterError::WriteConfig)?;
    let config_filename = config_filename.display().to_string();

    adminlog.writeln("execute_ci: run ambient-driver")?;
    let filename = projects_filename.display().to_string();
    let (exit, _) = runcmd(
        adminlog,
        dry_run,
        &[
            "env",
            "AMBIENT_LOG=debug",
            "ambient-driver",
            "--config",
            &config_filename,
            "run",
            "--force",
            "--projects",
            filename.as_str(),
        ],
        Path::new("."),
    )?;

    let (exit2, run_log_raw) = get_log(adminlog, dry_run, &project_name, &config_filename)?;

    if exit2 == 0 {
        Ok((exit, Some(run_log_raw)))
    } else {
        Ok((exit, None))
    }
}

fn get_log(
    adminlog: &mut AdminLog,
    dry_run: bool,
    project_name: &str,
    config_filename: &str,
) -> Result<(i32, Vec<u8>), AdapterError> {
    runcmd(
        adminlog,
        dry_run,
        &[
            "ambient-driver",
            "--config",
            config_filename,
            "log",
            project_name,
        ],
        Path::new("."),
    )
}

// Run a program.
pub fn runcmd(
    adminlog: &mut AdminLog,
    dry_run: bool,
    argv: &[&str],
    cwd: &Path,
) -> Result<(i32, Vec<u8>), AdapterError> {
    assert!(!argv.is_empty());

    if dry_run {
        adminlog.writeln(&format!("runcmd: pretend to run: argv={argv:?}"))?;
        return Ok((0, vec![]));
    }

    adminlog.writeln(&format!("runcmd: argv={argv:?}"))?;
    let output = Command::new("bash")
        .arg("-c")
        .arg(r#""$@" 2>&1"#)
        .arg("--")
        .args(argv)
        .current_dir(cwd)
        .output()
        .map_err(|err| AdapterError::Invoke("bash", err))?;

    let exit = output.status.code().unwrap_or(NO_EXIT);

    if exit != 0 {
        adminlog.writeln(&format!("runcmd: exit={exit}"))?;
        indented(adminlog, "stdout", &output.stdout);
        indented(adminlog, "stderr", &output.stderr);
    }

    Ok((exit, output.stdout))
}

fn indented(adminlog: &mut AdminLog, msg: &str, bytes: &[u8]) {
    if !bytes.is_empty() {
        adminlog.writeln(&format!("{msg}:")).ok();
        let text = String::from_utf8_lossy(bytes);
        for line in text.lines() {
            adminlog.writeln(&format!("    {line}")).ok();
        }
    }
}
